/** @file  VDIF_Frame.hpp 
 *  A interface Class for VDIF Frame.
 * @Version $Rev: 198 $
 * @Author M.Sekido
 * @Date  $Date: 2016-08-10 10:59:09 +0900 (Wed, 10 Aug 2016) $
 * $Id: VDIF_Frame.hpp 198 2016-08-10 01:59:09Z vlbi $ 
 */


#ifndef __VDIF_Frame_HPP__
#define __VDIF_Frame_HPP__

#include "VDIF_Header.hpp"
#include "VLBI_Frame.hpp"
#include <cstdio>
#include <cstring>

extern "C"{
#include <errno.h>
}

/// Data Acquisition System Types
//enum DAS_t {K5VSSP, ADS2000, ADS1000, ADS3000, ADS3000plus, MK5B, Undef};

/** 
 * Frame size: 
 * VDIF Frame size must be multiple of 8bytes.
 * - IP header size = 24 bytes;
 * - UDP header size = 8 bytes
 * - SUDP  header size = 8byte
 * - VDIF header size = 32 bytes;
 * - Available space for application = 1500-32 = 1468
 * - Application Transport Protocol Header (ATPH) = 8 bytes
 * - Multiple of 8byte for VLBI_frame = 1456 at maximum.
 */

const int DEFAULT_VDIF_FRAMESIZE = 1456; 

struct ReturnFrameInfo{
  enum ReadState_t {OK,NG,JUMP};
  int frame_intvl; // Number of frame interval to the next frame 
  //  bool no_jump;    // True if no jump in reading
  uint32_t  nframe; // Number of frame stored.
  string   msg;    // Message;
  ReadState_t state; // Reading Status 
public:
  ReturnFrameInfo():frame_intvl(0),nframe(0),state(OK){;}
  ~ReturnFrameInfo(){;}
  void Set_NG(string& m){state=NG; msg=m;}
  void Set_OK(const uint32_t nfrm){nframe=nfrm; state=OK;}
  void Set_Jump(const int frmcnt, const int jmp ){
    frame_intvl=jmp; state=JUMP; nframe = frmcnt;}
  void Set_Message(const string& m){msg = m;}
  ReadState_t State()const{ return state;}
};

//========== VDIF ===from Here  ====================
/** VDIF Frame class.
 * Secceeding from VLBI_Frame class 
 */
class VDIF_Frame: public VLBI_Frame{
  //  K5_VSI *k5vsi_dev;  //< K5_VSI Device  
  VDIF_Hdr _vdif_hdr;   ///< VDIF Header prototype  
  //  vector <uint8_t> packet; //< packet Frame including header
  //  time_t ref_epoch0; ///< Reference Epoch 0
  //  Time_dat _ref_epoch0; ///< Reference Epoch 
  FILE *_fp;  ///< File pointer for reading VDIF file.
  vector <uint8_t> _workbuf; ///< Buffer for reading data.
  ///uint32_t  nframe_sec; ///< Number of frames in 1 sec. comment out 2014/5/22 miya
  /// test-read flag. -1:Initial, 0:Checked, n>0, number of Hdr jumps.
  int   _hdr_tested; 
public:
  // Reading mode  (vdif_chk and vdif_speana).
  enum Read_t {DEEP, QUICK};
    //  VDIF_Frame(): k5vsi_dev(0){}
    //  VDIF_Frame():ref_epoch0(0){;}
  VDIF_Frame(); ///< Constructor 
  ~VDIF_Frame(){ }
  /// Returns VLBI-Frame type in number 
  const VLBI_Frame_t Type()const{return VDIF_frm;}
  /// Returns VLBI-Frame type in string
  string TypeName(){return string("VDIF");}
  /** Set Framesize (byte) */
  void Set_Framesize(const uint32_t sz){
    framesize =sz; _vdif_hdr.Set_FrameLen(sz);}
  /** Setting frame size by number of divisions/sec. */
  int32_t Set_FrameDiv(const int n=10);
  //  int Decompose_frame(const uint8_t* buf, uint32_t size ){return 0;}
  /** Set frame header and payload data (size byte) from data src 
   * \param ptk output pointer address for packet
   * \param src pointer address of data source
   * \param size data size to be copied from data source
   * \return return total data size of the packet.
   */
  int Make_frame( uint8_t* pkt, const uint8_t* src, const int size );
  /// Framesize (bytes) including header 
  //  uint32_t Framesize()const{return packet.size();};
  /// Returns Headersize (bytes)
  int Headsize()const{return VDIF_Hdr_size;}
  /// Set K5VSI  device module
  //  void Set_K5VSI( K5_VSI *k){k5vsi_dev = k;}
  /** Set Sampling parameters, s:(Hz),  return data rate (Bytes/sec)  */
  int32_t Set_Sparam(const int n, const int q, const uint32_t Hz);
  /// Set Initial Epoch
  //  void Set_Epoch0(const time_t t1){ _ref_epoch0 = _vdif_hdr.Set_Epoch0(t1);}
  /// Set Epoch to the header
  void Set_Epoch(const time_t t1);
  /// Set Epoch to the header
  void Set_Epoch(const Time_dat& td);
  /// Set Header to the packet 
  int  Set_header( uint8_t* buf, const int ifrm=-1);
  /// Set Station ID 
  void Set_StnID(const char* str){_vdif_hdr.Set_Stnid(str);}
  /// Reset sequential number of header
  void Reset_header(){_vdif_hdr.Frame_Number=0;} 
  const int Set_DasInfo(const DAS_Info& ds);
  /// Set Headr and DAS parameter to the packet.
  void Set_Head_par(const time_t t1, DAS_Info* dasinf=0);
  /// Set Station-name and Station-ID
  void Set_StnameID(const char* name, const char* id=0);
  /// Retrun Stname;
  void Get_Stname(string & name){_vdif_hdr.exud_ver1.Get_Stname(name);}
  /// Retrun Station ID
  void Get_StnID(string & id){_vdif_hdr.Get_Stnid(id);}
  /// Open for reading;
  const bool ROpen(string fname){
    _fp = fopen( fname.c_str(), "r"); 
    if (_fp == NULL){
      fprintf(stderr,"E VDIF_Frame:ROpen():%s \n",strerror(errno) );
      return false; }
    else return true;
  }
  VDIF_Hdr& Hdr(){ return _vdif_hdr;}
  /// Return Tested status
  const int Tested()const {return _hdr_tested;}
  /**! Read data in unit of seconds
   *  \param sec Time duration for reading data.
   *  \return number of frames, could be read.
   **/
  const long Read_sec(vector <uint8_t>& rawbuf, const int sec=1);
  /**! Read data in unit of frames
   *  \param hdarr header array to store the header data for return
   *  \param rawbuf data buffer for return
   *  \param frame Number of freq to request for input
   *  \return number of read frames. 
   **/
  const ReturnFrameInfo Read_frame(vector <VDIF_Hdr>& hdrarr, 
				   vector <uint8_t>& rawbuf,
				   const int frame=1);
  /**! Read all header of every second tics and returns its array
   *  \return Total Frame number in the data.
   **/
  const long Read_every_sec_hdr(vector <VDIF_Hdr>& hdrarr, 
				const size_t nsec=0, const Read_t mode=QUICK);
  /**! Read all header  and returns its array
   *  \return Total Frame number in the data.
   **/
  const long Read_every_hdr(vector <VDIF_Hdr>& hdrarr);
  /**  * Rewind file  * */
  const void Rewind(void ){rewind(_fp);}
  /**  * Skip reading Frame in the file 
   * \param ifrm Number of frame to skip(+/-)
   */
  const bool SeekFrames(const int ifrm );

  const bool EndofFile_Yes(void)const {
    bool ret = (feof(_fp) != 0);
    return ret;}
  /**! Return Number of Frames/sec  *   */
  const uint32_t Nframe_sec(void);
  /** Parameter for NPDF  */
  void Set_RFlb(const uint32_t rf_khz){_vdif_hdr.Set_RFlb(rf_khz);}
  uint32_t RFlb(void){
    VDIF_EUD_npdf *pd = static_cast <VDIF_EUD_npdf *> (& _vdif_hdr.exud_ver1); 
    return  pd->RFlb();}
  void Set_Nframe_sec(const uint32_t n);
};


inline const bool VDIF_Frame::SeekFrames(const int ifrm ){ 
  int64_t skip_size = ifrm * framesize; 
  bool ret=   fseek(_fp,skip_size,SEEK_CUR) == 0;
  return ret;
}

inline void VDIF_Frame::Set_StnameID(const char* name, const char* id){
  if(_vdif_hdr.EDV() == EDV_NICT) _vdif_hdr.exud_ver1.Set_stname(name);
  if(id!=0) _vdif_hdr.Set_Stnid(id);
}

inline  const int VDIF_Frame::Set_DasInfo(const DAS_Info& ds){
  das_inf = ds;
  _vdif_hdr.Set_Nch(ds.Nch());
  _vdif_hdr.Set_Qbit(ds.Qbit());
  _vdif_hdr.Set_Srate(ds.Srate()/1000);
  framesize = Headsize() + ds.Srate()/8 * ds.Qbit()* ds.Nch(); // add sekido 2013/12/20
  printf("VDIF_Frame %d\n",framesize );
  long bytesec = ds.Bytes_sec();
  int pldsize = framesize - Headsize();
  if(bytesec % pldsize != 0){
    Logsystem::info_logger()->Tstamp() 
      << "I VDIF_Frame::Set_DasInfo(): mode("<< ds.Mode() 
      << ")Bytesec(" << bytesec << ") modulo pldsize(" << pldsize 
      << ")!=0.";
    int pld = pldsize;
    while(bytesec % pld !=0) pld--; // Determining Frame size
    Set_Framesize(pld + Headsize());
    *Logsystem::info_logger() <<  "Adjusting. pldsize to " << pld
			      << " byte.\n";
  }
  Set_Nframe_sec(Nframe_sec()); // add miya 2015/1/27
  printf("VDIF_Frame %d\n",framesize );
  return 0;
}



inline  void VDIF_Frame::Set_Head_par(const time_t t1, DAS_Info* dasinf){
    VLBI_Frame::Set_Head_par(t1,dasinf);
    if(dasinf !=0){
	_vdif_hdr.Set_Nch(dasinf->Nch());
	_vdif_hdr.Set_Qbit(dasinf->Qbit());
	_vdif_hdr.Set_Srate(dasinf->Srate()/1000); // input in kHz
    }
}

inline int VDIF_Frame::Make_frame( uint8_t* pkt, const uint8_t* src, 
				   const int tobe_pkt){
  //  Set_header(pkt);

  //  if(ifrm >=0)  vdif_hdr.Frame_Number  = ifrm;  
  memcpy((void*)pkt, (void*)&_vdif_hdr, VDIF_Hdr_size);
  _vdif_hdr.Frame_Number++; // Count up

  pkt += VDIF_Hdr_size;
  memcpy( pkt, src, tobe_pkt );
  return tobe_pkt + VDIF_Hdr_size;
}

inline  void VDIF_Frame::Set_Epoch(const time_t t1){
  VLBI_Frame::Set_Epoch(t1);
  _vdif_hdr.Set_Epoch(data_epoch);
}

inline void VDIF_Frame::Set_Epoch(const Time_dat& t1){
    data_epoch=t1;
    _vdif_hdr.Set_Epoch(data_epoch);
}


/** Set Sampling parameters, s:(Hz),  return data rate (Bytes/sec)  */
inline 
int32_t VDIF_Frame::Set_Sparam(const int n, const int q, const uint32_t Hz){
    _vdif_hdr.Set_Sparam( n, q, Hz ); 
    das_inf.Set_Param( n, q, Hz );
    //  framesize=vdif_hdr.Framesize();
    return das_inf.Speed()*4;
}


inline int VDIF_Frame::Set_header( uint8_t* buf, const int ifrm){

  if(ifrm >=0)  _vdif_hdr.Frame_Number  = ifrm;  
  memcpy((void*)buf, &_vdif_hdr, VDIF_Hdr_size);
  //memmove((void*)buf, (void*)&vdif_hdr, VDIF_Hdr_size);
  //strncpy((char*)buf, (char *)&vdif_hdr,VDIF_Hdr_size);

  _vdif_hdr.Frame_Number++; // Count up
  
  return VDIF_Hdr_size;
}

//========== VDIF ===to   Here  ====================


#endif
