
#ifndef __VDIF_HEADER__
#define __VDIF_HEADER__
/** VDIF Class Header 
 * @Version $Rev: 202 $
 * @Author M.Sekido
 * @Date  $Date: 2016-09-29 22:22:30 +0900 (Thu, 29 Sep 2016) $
 * $Id: VDIF_Header.hpp 202 2016-09-29 13:22:30Z vlbi $ 
 */

//#include "VLBI_Header.hpp"
#include  <iostream>
#include  <sstream>
#include  <ctime>
#include  <cstring>
#include  <cstdio>
#include  <cstdlib>
#include  <string>
#include  <vector>
#include "time_lib.hpp"
using namespace std;

extern "C" {
#include  <stdint.h>
#include  <error.h>
#include  <errno.h>
}

const int VDIF_Hdr_size = 32;

//========== VDIF ===from Here  ====================

/// VDIF Sync patter of EDV1(NICT) Version
const uint32_t VDIF_NICT_SYNC = 0xACABFEED;

enum sample_rate_t {KSPS, MSPS};
const  uint8_t EDV_STD=0, EDV_NICT=0x01, EDV_NON=0, EDV_NRAO=0x0c, EDV_HAYSTACK=0xab, EDV_NPDF=128; // EDV Number designation see http://vlbi.org/vdif/


struct U_npdf{
  unsigned _rflb_khz:24;
  unsigned char buf[5];
};

/// VDIF Extended user data ver1(NICT)
struct VDIF_EUD_ver1{
  unsigned srate:23;    ///< Sampling Rate (ksps/Msps)
  unsigned sr_unit:1;   ///< Sampling Rate unit (0:ksps,1:Msps)
  uint8_t ext_dat_ver;  ///< Extended User data version Number
  uint32_t sync;        ///< Sync pattern 'ACABFEED'
  char    DAST_name[8]; ///< DAS/Station name 
public:
  /// Default Constructor of VDIF_EUD_nict1
  VDIF_EUD_ver1(): srate(32000),sr_unit(KSPS),ext_dat_ver(EDV_NICT), 
		    sync(VDIF_NICT_SYNC){
    memset(DAST_name,0,8);}
  /// Initialize the header
  void Init(const char* name="KASHIM34");
  /// Set Sampling Rate (kHz/MHz)
  void Set_Srate(unsigned long sr, const sample_rate_t Msps_unit=KSPS){
    srate=sr; sr_unit=Msps_unit;}
  /// Set Station name
  void Set_stname(const char *name){strncpy(DAST_name,name,8);}
  /// Returns Station name in string 
  void Get_Stname(string& name)const;
  void Set_RFlb(const unsigned long khz){
    U_npdf u;
    u._rflb_khz = khz;  
    memcpy(DAST_name,&u,8); 
  }
};

inline  void VDIF_EUD_ver1::Init(const char* name){
  ext_dat_ver = EDV_NICT; // NICT
  sync = VDIF_NICT_SYNC; 
  strncpy(DAST_name,name,8);
}

inline void VDIF_EUD_ver1::Get_Stname(string& name)const{
  char buf[10]; 
  strncpy(buf,DAST_name,8); buf[8]='\0';
  name = buf;
}

/// VDIF Extended user data ver128(NPDF)
struct VDIF_EUD_npdf:public VDIF_EUD_ver1 { ///< NICT Pulsar Data format
public:
  /// Set RF Lower boubd Freq. [kHz]
  void Set_RFlb(const unsigned long khz){
    U_npdf u;
    u._rflb_khz = khz;  
    memcpy(DAST_name,&u,8); 
  }
  /// Returns RF Lower boubd Freq. [kHz]
  uint32_t RFlb(void)const{ U_npdf *u = (U_npdf*)DAST_name;  return (uint32_t)u->_rflb_khz;}
};

struct VDIF_Hdr{
  /** Print Information level **/
  enum Prnt_t {ALL=1, CONST=20, VAR=21, ONELINE=22,ALL_ONE=2 };
  enum Data_t {Real_Sample, Complex_Sample};
  /** Seconds from reference Epoch **/
  unsigned  Sec_from_Ref:30;
  /** Flag for legacy compatibility, 0:32 Bytes header, 1:16 Bytes header*/
  unsigned  Legacy_flag:1;
  /** Invalid flag 1: Invalid, 0: Valid */
  unsigned  Invalid_flag:1;
  /** Data Frame Number */
  unsigned  Frame_Number:24;
  /** Reference Epoch half year counter since 2000.0 */
  unsigned  Ref_Epoch_hy_counter:6;
  /** unused  */
  unsigned  unused:2;
  /** Data Frame length -1, unit: 8 bytes  */
  unsigned  Frame_Length:24;
  /** Log2(nch)*/
  unsigned  Lg_Nch:5;
  /** VDIF Version Number */
  unsigned  Version_No:3;
  /** Station ID*/
  unsigned char Station_ID[2];
  /** Thread ID*/
  unsigned  Thread_Id:10;
  /** Quantization Bit/sample -1 */
  unsigned  Qbit_1:5;
  /** Comlex data flag. 1:Complex, 0: Real data */
  unsigned  Complex_data_flag:1;
  /** Extended user data 1 */
  VDIF_EUD_ver1 exud_ver1;
public:
  /// Constructor 
  VDIF_Hdr(){
    memset(this,0,sizeof(VDIF_Hdr)); 
    Version_No=1;Thread_Id=0;Qbit_1=1; 
    strncpy((char*)Station_ID,"kb",2);
    exud_ver1.Init();
  }
  ~VDIF_Hdr(){;  }  ///< Destructor 
  const bool Valid() const { return ( Invalid_flag ? false : true);}
  /// Returns Total second from Reference Epoch 
  uint32_t Sec_Ref()const{   return  (uint32_t)Sec_from_Ref;    }
  int32_t  iFrame()const{return Frame_Number;} ///< Returns Frame Number 
  int Nch() const {return (1 << Lg_Nch); }  ///< Returns Nch
  int Qbit()const {return Qbit_1+1;}  ///< returns Qunatization bits
  /// Sampling rate (Hz)
  uint32_t Srate()const{return Srate_ksps()*1000;}
  //  time_t Ref_Epoch() const;  ///< Reference Epoch
  Time_dat Ref_Epoch() const;  ///< Reference Epoch
  /**! Printing Header data
   * \param fmt pattern 1,2,20,21
   */
  void Print_hd(FILE *fp=stdout, const Prnt_t fmt=ALL){
    fprintf(fp,"%s\n", Str_hd(fmt).c_str());
  };
  /**! Header data printing to string 
   * \param fmt pattern 1,2,20,21
   * \return string 
   */
  string Str_hd(const Prnt_t fmt=VDIF_Hdr::ALL);
  /// return framesize including header (bytes)
  int Framesize(){long fs = (Frame_Length)*8; 
    return ( fs < VDIF_Hdr_size ? VDIF_Hdr_size : fs); }
  /**!
   * \return header size (bytes)
   */
  int Headsize()const {return (VDIF_Hdr_size);}
  /**!
   * \return Time_dat data of the epoch
   */
  Time_dat Epoch()const;
  /// Return integer MJD of the epoch
  uint32_t Epoch_imjd(){return (uint32_t)(floor(Epoch().Mjd()));}
  /// Returns Second of the day
  double Epoch_sod() {return Epoch().Sod();}  ///< return sec of the day (sec)  
  /// Return header type
  //  VLBI_Hdr_t Header_type()const {return VDIF_Header_Type;}
  bool HD_Match() const{return (exud_ver1.sync == VDIF_NICT_SYNC ? true: false);}
  /// Setting  Nch 
  void Set_Nch(const int nch);
  /// Setting Station ID
  void Set_Stnid(const char *id){strncpy((char*)Station_ID,(char*)id,2);}
  /// Setting Quantization bit
  void Set_Qbit(const int nbit){Qbit_1 = nbit-1;}
  /// Setting Sampling Freq (ksps);
  void Set_Srate(const uint32_t ksps);
  /// Return Sampling rate in ksps
  uint32_t Srate_ksps()const; 
  /// Setting Framesize (Bytes)
  void Set_FrameLen(const uint32_t fbyte){Frame_Length = fbyte/8;}
  /// Setting Sampling Parameters (nch,qbit,srate) a nd Framesize at a time
  int  Set_Sparam(const int nch, const int qbit, const uint32_t Hz);
  /// Setup Epoch parametes;
  void Set_Epoch(const Time_dat &t1);
  /// Set Initial Epoch. 
  /// When t1 is given, nearst ealier Epoch is automatically set.
  Time_dat Set_Epoch0(const time_t t1);
  //  void Set_Epoch(const time_t t1, const time_t t0_ref);
  /// Set Frame number 
  void Set_iFrame(const uint32_t iframe){ Frame_Number = iframe;}
  /// Counting up Frame Number ;
  void iFrame_plus(){Frame_Number++;}
  /// Returns Total Data rate in word(4 bytes)/sec
  int32_t Speed()const { 
    if(HD_Match()) return (Srate()/32 * Nch() * Qbit()); else return -1;
  }
  /// Returns Data Type (Real/Complex)
  string Dtype(void)const {string ret = (Complex_data_flag == 1 ? "Cmplx" :"Real"); 
    return ret;}
  /// Get Station ID
  void Get_Stnid(string & id) const{ 
    char buf[3]; strncpy(buf,(char*)Station_ID,2); buf[2]='\0'; id=buf;}
  const string Stnid(void) const{ 
    char buf[5]; strncpy(buf,(char*)Station_ID,2); buf[2]='\0';
    string ss(buf);
    return ss; }
  const string Stname() const{
    string str="N/A"; 
    if(exud_ver1.ext_dat_ver == EDV_NICT)  exud_ver1.Get_Stname(str);
    return str; }
  const string Version(void){
    char str[256];
    sprintf(str, "Ver.%d.%d", Version_No,exud_ver1.ext_dat_ver);
    string ss = str;
    return ss;
  }
  const int ThreadID(void) {  return  Thread_Id; }
  /// Returns Extended User Data Version 
  const uint8_t EDV() const {return exud_ver1.ext_dat_ver;}
  /// Set EUD Version
  void Set_edv(const uint8_t edv){ exud_ver1.ext_dat_ver = edv;}
  void Set_RFlb(const unsigned long khz);
  const Data_t Data_Type(){return (Complex_data_flag == 1 ? Complex_Sample : Real_Sample);}
};

inline void VDIF_Hdr::Set_RFlb(const unsigned long khz){
  exud_ver1.Set_RFlb(khz);
  Set_edv(EDV_NPDF);
}

inline  void VDIF_Hdr::Set_Srate(const uint32_t ksps){
  if(ksps <1000000){
    exud_ver1.Set_Srate(ksps, KSPS);
  }else{
    exud_ver1.Set_Srate(ksps/1000,MSPS);
  }
}

inline  uint32_t VDIF_Hdr::Srate_ksps()const{
  switch(exud_ver1.ext_dat_ver){
  case EDV_NICT:
  case EDV_NRAO:
    if(exud_ver1.sr_unit){ // Msps
      return exud_ver1.srate*1000;
    }else{ // ksps
      return exud_ver1.srate;
    }
    break;
  case EDV_HAYSTACK:
  default:
    return 0;
    break;
  }
  
}

inline  void VDIF_Hdr::Set_Nch(const int nch){
  int lch =0;
  int ich=1;
  while( (nch & ich) == 0){ ich <<= 1; lch++;}
  Lg_Nch = lch;
}

inline Time_dat VDIF_Hdr::Epoch()const{
  Time_dat t0;
  t0 =  Ref_Epoch();
  t0 += Sec_from_Ref;
  return t0;
}


enum VDIF_Ext_t {No_ext, NICT_ext };
/*!  Header Finder from buffer data
 *  \param buf data buffer pointer for input
 *  \param nbyte buffer size in byte for input.
 *  \param offset position of detected header for output
 *  \param hdr pointer of the header container for output
 ***/

const bool VDIF_Hdr_find(const uint8_t*  buf, const uint32_t nbyte,
			 long &offset, VDIF_Hdr* hdr);

/*!  Header Finder from a file.  
 *  \param fp  File pointer for input
 *  \param offset position of detected header for output
 *  \param hdr pointer of the header container for output
 *  File pointer is set to the top of header.
 ***/
const bool VDIF_Hdr_find(FILE *fp, long &offset, VDIF_Hdr* hdr);

// /// @class VDIF_Header is header class of VDIF
// class VDIF_Header: public VLBI_Header{
//   VDIF_Hdr* pt_head;
// public:
//   VDIF_Header():VLBI_Header(VDIF_Hdr_size),pt_head(0){;} /// < Default Constructor 
//   /// Header finder
//   /// @return  Number of bytes to find the header, or -1: for Error
//   int32_t Hd_find(const uint8_t*  buf, const uint32_t nbyte);
//   /// Print Header Information
// };

//========== VDIF ===to   Here  ====================

#endif
